/*
  This file is part of TALER
  Copyright (C) 2024-2026 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2.1,
  or (at your option) any later version.

  TALER is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with TALER; see the file COPYING.LGPL.
  If not, see <http://www.gnu.org/licenses/>
*/
/**
 * @file src/lib/merchant_api_post-private-donau.c
 * @brief Implementation of the POST /private/donau request
 * @author Christian Grothoff
 */
#include "platform.h"
#include <curl/curl.h>
#include <jansson.h>
#include <microhttpd.h> /* just for HTTP status codes */
#include <gnunet/gnunet_util_lib.h>
#include <gnunet/gnunet_curl_lib.h>
#include <taler/merchant/post-private-donau.h>
#include "merchant_api_curl_defaults.h"
#include "merchant_api_common.h"
#include <taler/taler_json_lib.h>
#include <taler/taler_curl_lib.h>


/**
 * Handle for a POST /private/donau operation.
 */
struct TALER_MERCHANT_PostPrivateDonauHandle
{
  /**
   * Base URL of the merchant backend.
   */
  char *base_url;

  /**
   * The full URL for this request.
   */
  char *url;

  /**
   * Handle for the request.
   */
  struct GNUNET_CURL_Job *job;

  /**
   * Function to call with the result.
   */
  TALER_MERCHANT_PostPrivateDonauCallback cb;

  /**
   * Closure for @a cb.
   */
  TALER_MERCHANT_POST_PRIVATE_DONAU_RESULT_CLOSURE *cb_cls;

  /**
   * Reference to the execution context.
   */
  struct GNUNET_CURL_Context *ctx;

  /**
   * Minor context that holds body and headers.
   */
  struct TALER_CURL_PostContext post_ctx;

  /**
   * Charity information to register.
   */
  struct TALER_MERCHANT_Charity charity;

  /**
   * Charity URL (owned copy).
   */
  char *charity_url;
};


/**
 * Function called when we're done processing the
 * HTTP POST /private/donau request.
 *
 * @param cls the `struct TALER_MERCHANT_PostPrivateDonauHandle`
 * @param response_code HTTP response code, 0 on error
 * @param response response body, NULL if not in JSON
 */
static void
handle_post_donau_finished (void *cls,
                            long response_code,
                            const void *response)
{
  struct TALER_MERCHANT_PostPrivateDonauHandle *ppdh = cls;
  const json_t *json = response;
  struct TALER_MERCHANT_PostPrivateDonauResponse pdr = {
    .hr.http_status = (unsigned int) response_code,
    .hr.reply = json
  };

  ppdh->job = NULL;
  GNUNET_log (GNUNET_ERROR_TYPE_INFO,
              "POST /private/donau completed with response code %u\n",
              (unsigned int) response_code);
  switch (response_code)
  {
  case 0:
    pdr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
    break;
  case MHD_HTTP_ACCEPTED:
    if (GNUNET_OK !=
        TALER_MERCHANT_parse_mfa_challenge_response_ (
          json,
          &pdr.details.accepted))
    {
      GNUNET_break_op (0);
      pdr.hr.http_status = 0;
      pdr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
    }
    break;
  case MHD_HTTP_NO_CONTENT:
    break;
  case MHD_HTTP_BAD_REQUEST:
    pdr.hr.ec = TALER_JSON_get_error_code (json);
    pdr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  case MHD_HTTP_CONFLICT:
    pdr.hr.ec = TALER_JSON_get_error_code (json);
    pdr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  case MHD_HTTP_INTERNAL_SERVER_ERROR:
    pdr.hr.ec = TALER_JSON_get_error_code (json);
    pdr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  case MHD_HTTP_BAD_GATEWAY:
    pdr.hr.ec = TALER_JSON_get_error_code (json);
    pdr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  default:
    TALER_MERCHANT_parse_error_details_ (json,
                                         response_code,
                                         &pdr.hr);
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u/%d\n",
                (unsigned int) response_code,
                (int) pdr.hr.ec);
    GNUNET_break_op (0);
    break;
  }
  ppdh->cb (ppdh->cb_cls,
            &pdr);
  if (MHD_HTTP_ACCEPTED == response_code)
    TALER_MERCHANT_mfa_challenge_response_free (
      &pdr.details.accepted);
  TALER_MERCHANT_post_private_donau_cancel (ppdh);
}


struct TALER_MERCHANT_PostPrivateDonauHandle *
TALER_MERCHANT_post_private_donau_create (
  struct GNUNET_CURL_Context *ctx,
  const char *url,
  const struct TALER_MERCHANT_Charity *charity)
{
  struct TALER_MERCHANT_PostPrivateDonauHandle *ppdh;

  ppdh = GNUNET_new (struct TALER_MERCHANT_PostPrivateDonauHandle);
  ppdh->ctx = ctx;
  ppdh->base_url = GNUNET_strdup (url);
  ppdh->charity_url = GNUNET_strdup (charity->charity_url);
  ppdh->charity.charity_url = ppdh->charity_url;
  ppdh->charity.charity_id = charity->charity_id;
  return ppdh;
}


enum GNUNET_GenericReturnValue
TALER_MERCHANT_post_private_donau_set_options_ (
  struct TALER_MERCHANT_PostPrivateDonauHandle *ppdh,
  unsigned int num_options,
  const struct TALER_MERCHANT_PostPrivateDonauOptionValue *options)
{
  for (unsigned int i = 0; i < num_options; i++)
  {
    switch (options[i].option)
    {
    case TALER_MERCHANT_POST_PRIVATE_DONAU_OPTION_END:
      return GNUNET_OK;
    default:
      GNUNET_break (0);
      return GNUNET_SYSERR;
    }
  }
  return GNUNET_OK;
}


enum TALER_ErrorCode
TALER_MERCHANT_post_private_donau_start (
  struct TALER_MERCHANT_PostPrivateDonauHandle *ppdh,
  TALER_MERCHANT_PostPrivateDonauCallback cb,
  TALER_MERCHANT_POST_PRIVATE_DONAU_RESULT_CLOSURE *cb_cls)
{
  json_t *req_obj;
  CURL *eh;

  ppdh->cb = cb;
  ppdh->cb_cls = cb_cls;
  ppdh->url = TALER_url_join (ppdh->base_url,
                              "private/donau",
                              NULL);
  if (NULL == ppdh->url)
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  req_obj = GNUNET_JSON_PACK (
    GNUNET_JSON_pack_string ("donau_url",
                             ppdh->charity.charity_url),
    GNUNET_JSON_pack_uint64 ("charity_id",
                             ppdh->charity.charity_id)
    );
  eh = TALER_MERCHANT_curl_easy_get_ (ppdh->url);
  if ( (NULL == eh) ||
       (GNUNET_OK !=
        TALER_curl_easy_post (&ppdh->post_ctx,
                              eh,
                              req_obj)) )
  {
    GNUNET_break (0);
    json_decref (req_obj);
    if (NULL != eh)
      curl_easy_cleanup (eh);
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  }
  json_decref (req_obj);
  ppdh->job = GNUNET_CURL_job_add2 (ppdh->ctx,
                                    eh,
                                    ppdh->post_ctx.headers,
                                    &handle_post_donau_finished,
                                    ppdh);
  if (NULL == ppdh->job)
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  return TALER_EC_NONE;
}


void
TALER_MERCHANT_post_private_donau_cancel (
  struct TALER_MERCHANT_PostPrivateDonauHandle *ppdh)
{
  if (NULL != ppdh->job)
  {
    GNUNET_CURL_job_cancel (ppdh->job);
    ppdh->job = NULL;
  }
  TALER_curl_easy_post_finished (&ppdh->post_ctx);
  GNUNET_free (ppdh->charity_url);
  GNUNET_free (ppdh->url);
  GNUNET_free (ppdh->base_url);
  GNUNET_free (ppdh);
}


/* end of merchant_api_post-private-donau.c */
